import os
import importlib

from tests.utests.common import (
    lakectl_test_config_context,
    lakectl_no_config_context,
    env_var_context, TEST_SERVER,
    TEST_ACCESS_KEY_ID,
    TEST_SECRET_ACCESS_KEY,
    TEST_ENDPOINT_PATH
)

from lakefs import config as client_config

TEST_CONFIG_KWARGS: dict[str, str] = {
    "username": "my_username",
    "password": "my_password",
    "host": "http://my_host",
    "access_token": "my_jwt_token"
}


class TestClient:
    def test_client_no_config(self, monkeypatch):
        with lakectl_no_config_context(monkeypatch) as client:
            client.DEFAULT_CLIENT = None
            client = importlib.reload(client)
            assert client.DEFAULT_CLIENT is None

    def test_client_no_kwargs(self, monkeypatch, tmp_path):
        with lakectl_test_config_context(monkeypatch, tmp_path) as client:
            assert client.DEFAULT_CLIENT is not None
            config = client.DEFAULT_CLIENT.config
            assert config.host == TEST_SERVER + TEST_ENDPOINT_PATH
            assert config.username == TEST_ACCESS_KEY_ID
            assert config.password == TEST_SECRET_ACCESS_KEY

            with env_var_context():
                # Create a new client with env vars
                env_endpoint = "http://env_endpoint/api/v1"
                env_secret = "test_env_secret"
                os.environ[client_config._LAKECTL_ENDPOINT_ENV] = env_endpoint
                os.environ[client_config._LAKECTL_SECRET_ACCESS_KEY_ENV] = env_secret
                clt = client.Client()
                assert clt is not client.DEFAULT_CLIENT
                config = clt.config
                assert config.host == env_endpoint
                assert config.username == TEST_ACCESS_KEY_ID
                assert config.password == env_secret

    def test_client_kwargs(self, monkeypatch, tmp_path):
        # Use lakectl yaml file and ensure it is not being read in case kwargs are provided
        with lakectl_test_config_context(monkeypatch, tmp_path) as client:
            clt = client.Client(**TEST_CONFIG_KWARGS)
            config = clt.config
            assert config.host == TEST_CONFIG_KWARGS["host"] + TEST_ENDPOINT_PATH
            assert config.username == TEST_CONFIG_KWARGS["username"]
            assert config.password == TEST_CONFIG_KWARGS["password"]
            assert config.access_token == TEST_CONFIG_KWARGS["access_token"]

    def test_client_init(self, monkeypatch, tmp_path):
        with lakectl_no_config_context(monkeypatch) as client:
            assert client.DEFAULT_CLIENT is None
            from lakefs.client import init
            init(**TEST_CONFIG_KWARGS)
            assert client.DEFAULT_CLIENT is not None
            config = client.DEFAULT_CLIENT.config
            assert config.host == TEST_CONFIG_KWARGS["host"] + TEST_ENDPOINT_PATH
            assert config.username == TEST_CONFIG_KWARGS["username"]
            assert config.password == TEST_CONFIG_KWARGS["password"]
            assert config.access_token == TEST_CONFIG_KWARGS["access_token"]



"""
Module containing lakeFS reference implementation
"""

from __future__ import annotations

from typing import Optional, Generator, Literal, List

import lakefs_sdk

from lakefs.client import Client, DEFAULT_CLIENT
from lakefs.exceptions import api_exception_handler
from lakefs.object import StoredObject
from lakefs.namedtuple import LenientNamedTuple
from lakefs.object_manager import ObjectManager


class Commit(LenientNamedTuple):
    """
    NamedTuple representing a lakeFS commit's properties
    """
    id: str
    parents: List[str]
    committer: str
    message: str
    creation_date: int
    meta_range_id: str
    metadata: Optional[dict[str, str]] = None


class Change(LenientNamedTuple):
    """
    NamedTuple representing a diff change between two refs in lakeFS
    """
    type: Literal["added", "removed", "changed", "conflict", "prefix_changed"]
    path: str
    path_type: Literal["common_prefix", "object"]
    size_bytes: Optional[int]


class Reference:
    """
    Class representing a reference in lakeFS.
    """
    _client: Client
    _repo_id: str
    _id: str
    _commit: Optional[Commit] = None

    def __init__(self, repo_id: str, ref_id: str, client: Optional[Client] = DEFAULT_CLIENT) -> None:
        self._client = client
        self._repo_id = repo_id
        self._id = ref_id

    @property
    def repo_id(self) -> str:
        """
        Return the repository id for this reference
        """
        return self._repo_id

    @property
    def id(self) -> str:
        """
        Returns the reference id
        """
        return self._id

    @property
    def objects(self) -> ObjectManager:
        """
        Returns a ObjectManager object for this reference
        """
        # TODO: Implement

    @staticmethod
    def _get_generator(func, *args, max_amount: Optional[int] = None, **kwargs):
        has_more = True
        with api_exception_handler():
            while has_more:
                page = func(*args, **kwargs)
                has_more = page.pagination.has_more
                kwargs["after"] = page.pagination.next_offset
                for res in page.results:
                    yield res

                    if max_amount is not None:
                        max_amount -= 1
                        if max_amount <= 0:
                            return

    def log(self, max_amount: Optional[int] = None, **kwargs) -> Generator[Commit]:
        """
        Returns a generator of commits starting with this reference id

        :param max_amount: (Optional) limits the amount of results to return from the server
        :param kwargs: Additional keyword arguments
        :raises:
            NotFoundException if reference by this id does not exist
            NotAuthorizedException if user is not authorized to perform this operation
            ServerException for any other errors
        """
        for res in self._get_generator(self._client.sdk_client.refs_api.log_commits,
                                       self._repo_id, self._id, max_amount=max_amount, **kwargs):
            yield Commit(**res.dict())

    def _get_commit(self):
        if self._commit is None:
            with api_exception_handler():
                commit = self._client.sdk_client.commits_api.get_commit(self._repo_id, self._id)
                self._commit = Commit(**commit.dict())
        return self._commit

    def metadata(self) -> dict[str, str]:
        """
        Return commit metadata for this reference id
        """
        return self._get_commit().metadata

    def commit_message(self) -> str:
        """
        Return commit message for this reference id
        """
        return self._get_commit().message

    def commit_id(self) -> str:
        """
        Return commit id for this reference id
        """
        return self._get_commit().id

    def diff(self,
             other_ref: str | Reference,
             max_amount: Optional[int] = None,
             after: Optional[str] = None,
             prefix: Optional[str] = None,
             delimiter: Optional[str] = None,
             **kwargs) -> Generator[Change]:
        """
        Returns a diff generator of changes between this reference and other_ref

        :param other_ref: The other ref to diff against
        :param max_amount: Stop showing changes after this amount
        :param after: Return items after this value
        :param prefix: Return items prefixed with this value
        :param delimiter: Group common prefixes by this delimiter
        :raises:
            NotFoundException if this reference or other_ref does not exist
            NotAuthorizedException if user is not authorized to perform this operation
            ServerException for any other errors
        """
        for diff in self._get_generator(self._client.sdk_client.refs_api.diff_refs,
                                        repository=self._repo_id,
                                        left_ref=self._id,
                                        right_ref=str(other_ref),
                                        after=after,
                                        max_amount=max_amount,
                                        prefix=prefix,
                                        delimiter=delimiter,
                                        **kwargs):
            yield Change(**diff.dict())

    def merge_into(self, destination_branch_id: str | Reference, **kwargs) -> str:
        """
        Merge this reference into destination branch

        :param destination_branch_id: The ID of the merge destination
        :return: The reference id of the merge commit
        :raises:
            NotFoundException if reference by this id does not exist, or branch doesn't exist
            NotAuthorizedException if user is not authorized to perform this operation
            ServerException for any other errors
        """
        with api_exception_handler():
            merge = lakefs_sdk.Merge(**kwargs)
            res = self._client.sdk_client.refs_api.merge_into_branch(self._repo_id,
                                                                     self._id,
                                                                     str(destination_branch_id),
                                                                     merge=merge)
            return res.reference

    def object(self, path: str) -> StoredObject:  # pylint: disable=C0103
        """
        Returns an Object class representing a lakeFS object with this repo id, reference id and path

        :param path: The object's path
        """
        return StoredObject(self._repo_id, self._id, path)

    def __str__(self) -> str:
        return self._id

    def __repr__(self):
        return f"lakefs://{self._repo_id}/{self._id}"



"""
Module containing lakeFS repository implementation
"""

from __future__ import annotations

from typing import Optional
import lakefs_sdk

from lakefs.tag import Tag
from lakefs.branch import Branch
from lakefs.branch_manager import BranchManager
from lakefs.client import Client, DEFAULT_CLIENT
from lakefs.exceptions import api_exception_handler, ConflictException, LakeFSException
from lakefs.namedtuple import LenientNamedTuple
from lakefs.reference import Reference
from lakefs.tag_manager import TagManager


class RepositoryProperties(LenientNamedTuple):
    """
    Represent a lakeFS repository's properties
    """
    id: str
    creation_date: int
    default_branch: str
    storage_namespace: str


class Repository:
    """
    Class representing a Repository in lakeFS.
    The Repository object provides the context for the other objects that are found in it.
    Access to these objects should be done from this class
    """
    _client: Client
    _id: str
    _properties: RepositoryProperties = None

    def __init__(self, repository_id: str, client: Optional[Client] = DEFAULT_CLIENT) -> None:
        self._id = repository_id
        self._client = client

    def create(self,
               storage_namespace: str,
               default_branch: str = "main",
               include_samples: bool = False,
               exist_ok: bool = False,
               **kwargs) -> Repository:
        """
        Create a new repository in lakeFS from this object

        :param storage_namespace: Repository's storage namespace
        :param default_branch: The default branch for the repository. If None, use server default name
        :param include_samples: Whether to include sample data in repository creation
        :param exist_ok: If False will throw an exception if a repository by this name already exists. Otherwise,
         return the existing repository without creating a new one
        :return: The lakeFS SDK object representing the repository
        :raises:
            NotAuthorizedException if user is not authorized to perform this operation
            ServerException for any other errors
        """
        repository_creation = lakefs_sdk.RepositoryCreation(name=self._id,
                                                            storage_namespace=storage_namespace,
                                                            default_branch=default_branch,
                                                            sample_data=include_samples)

        def handle_conflict(e: LakeFSException):
            if isinstance(e, ConflictException) and exist_ok:
                with api_exception_handler():
                    get_repo = self._client.sdk_client.repositories_api.get_repository(self._id)
                    self._properties = RepositoryProperties(**get_repo.dict())
                    return None
            return e

        with api_exception_handler(handle_conflict):
            repo = self._client.sdk_client.repositories_api.create_repository(repository_creation, **kwargs)
            self._properties = RepositoryProperties(**repo.dict())
        return self

    def delete(self) -> None:
        """
        Delete repository from lakeFS server

        :raises:
            NotFoundException if repository by this id does not exist
            NotAuthorizedException if user is not authorized to perform this operation
            ServerException for any other errors
        """
        with api_exception_handler():
            self._client.sdk_client.repositories_api.delete_repository(self._id)

    def branch(self, branch_id: str) -> Branch:
        """
        Return a branch object using the current repository id and client
        :param branch_id: name of the branch
        """
        return Branch(self._id, branch_id, self._client)

    def commit(self, commit_id: str) -> Reference:
        """
        Return a reference object using the current repository id and client
        :param commit_id: id of the commit reference
        """
        return Reference(self._id, commit_id, self._client)

    def ref(self, ref_id: str) -> Reference:
        """
        Return a reference object using the current repository id and client
        :param ref_id: branch name, commit id or tag id
        """
        return Reference(self._id, ref_id, self._client)

    def tag(self, tag_id: str) -> Tag:
        """
        Return a tag object using the current repository id and client
        :param tag_id: name of the tag
        """
        return Tag(self._id, tag_id, self._client)

    @property
    def metadata(self) -> dict[str, str]:
        """
        Returns the repository metadata
        """
        with api_exception_handler():
            return self._client.sdk_client.repositories_api.get_repository_metadata(repository=self._id)

    @property
    def branches(self) -> BranchManager:
        """
        Returns a BranchManager object for this repository
        """
        return BranchManager()

    @property
    def tags(self) -> TagManager:
        """
        Returns a TagManager object for this repository
        """
        return TagManager()

    @property
    def properties(self) -> RepositoryProperties:
        """
        Return the repositories properties object
        """
        if self._properties is None:
            with api_exception_handler():
                repo = self._client.sdk_client.repositories_api.get_repository(self._id)
                self._properties = RepositoryProperties(**repo.dict())
                return self._properties

        return self._properties



import io
from typing import get_args

import pytest

from tests.utests.common import expect_exception_context
from lakefs.exceptions import ObjectExistsException, InvalidRangeException, NotFoundException
from lakefs.object import WriteableObject, WriteModes, ReadModes


@pytest.mark.parametrize("pre_sign", (True, False), indirect=True)
def test_object_read_seek(setup_repo, pre_sign):
    clt, repo = setup_repo
    data = b"test_data"
    obj = WriteableObject(repository=repo.properties.id, reference="main", path="test_obj", client=clt).upload(
        data=data, pre_sign=pre_sign)

    with obj.reader() as fd:
        assert fd.read(2 * len(data)) == data
        fd.seek(2)

        assert fd.read(5) == data[2:7]

        assert fd.read() == data[7:]

        # This should raise an exception
        with expect_exception_context(InvalidRangeException):
            fd.read(1)

        fd.seek(0)
        for c in data:
            assert ord(fd.read(1)) == c
        # This should raise an exception
        with expect_exception_context(InvalidRangeException):
            fd.read(1)


def test_object_upload_exists(setup_repo):
    clt, repo = setup_repo
    data = b"test_data"
    obj = WriteableObject(repository=repo.properties.id, reference="main", path="test_obj", client=clt).upload(
        data=data)
    with expect_exception_context(ObjectExistsException):
        obj.upload(data="some_other_data", mode='xb')

    with obj.reader() as fd:
        assert fd.read() == data

    # Create - overwrite
    new_data = b"new_data"
    obj2 = obj.upload(data=new_data, mode='w')

    with obj.reader() as fd:
        assert fd.read() == new_data

    assert obj2 == obj


@pytest.mark.parametrize("w_mode", get_args(WriteModes))
@pytest.mark.parametrize("r_mode", get_args(ReadModes))
@pytest.mark.parametrize("pre_sign", (True, False), indirect=True)
def test_object_upload_read_different_params(setup_repo, w_mode, r_mode, pre_sign):
    clt, repo = setup_repo

    # urllib3 encodes TextIOBase to ISO-8859-1, data should contain symbols that can be encoded as such
    data = b'test \x48\x65\x6c\x6c\x6f\x20\x57\x6f\x72\x6c\x64\x21'
    obj = WriteableObject(repository=repo.properties.id, reference="main", path="test_obj", client=clt).upload(
        data=data, mode=w_mode, pre_sign=pre_sign)

    with obj.reader(mode=r_mode) as fd:
        res = fd.read()
        if 'b' in r_mode:
            assert res == data
        else:
            assert res.encode('utf-8') == data


def test_object_copy(setup_repo):
    clt, repo = setup_repo
    data = "test_data"
    obj = WriteableObject(repository=repo.properties.id, reference="main", path="test_obj", client=clt).upload(
        data=data, metadata={"foo": "bar"})

    copy_name = "copy_obj"
    copy = obj.copy("main", copy_name)
    obj_stat = obj.stat()
    copy_stat = copy.stat()

    assert copy != obj
    assert copy_stat.metadata == obj_stat.metadata
    assert copy_stat.path != obj_stat.path
    assert copy_stat.content_type == obj_stat.content_type
    assert copy_stat.physical_address != obj_stat.physical_address
    assert copy_stat.mtime >= obj_stat.mtime
    assert copy_stat.size_bytes == obj_stat.size_bytes
    assert copy_stat.checksum == obj_stat.checksum


def test_writer(setup_repo):
    _, repo = setup_repo
    obj = repo.branch("main").object("test_object")
    with obj.writer() as writer:
        with expect_exception_context(io.UnsupportedOperation):
            writer.seek(10)

        assert not writer.readable()
        assert writer.writable()

        writer.write("Hello")
        writer.write(" World!")

        # Check that the object does not exist in lakeFS before we close the writer
        with expect_exception_context(NotFoundException):
            obj.stat()

    assert obj.reader().read() == b"Hello World!"


@pytest.mark.parametrize("w_mode", get_args(WriteModes))
@pytest.mark.parametrize("r_mode", get_args(ReadModes))
@pytest.mark.parametrize("pre_sign", (True, False), indirect=True)
def test_writer_different_params(setup_repo, w_mode, r_mode, pre_sign):
    _, repo = setup_repo
    obj = repo.branch("main").object("test_object")
    writer = obj.writer(mode=w_mode)

    data = [
        "The quick brown fox jumps over the lazy dog"
        "Pack my box with five dozen liquor jugs"
        "Sphinx of black quartz, judge my vow" * 1000
    ]

    pos = 0
    for part in data:
        pos += len(part)
        writer.write(part)
        assert writer.tell() == pos

    # Check that the object does not exist in lakeFS before we close the writer
    with expect_exception_context(NotFoundException):
        obj.stat()

    writer.content_type = "text/plain"
    writer.metadata = {"foo": "bar"}
    writer.pre_sign = pre_sign

    writer.close()

    stats = obj.stat()
    assert stats.metadata == {"foo": "bar"}
    assert stats.content_type == "text/plain"
    assert stats.path == "test_object"
    assert stats.size_bytes == pos

    expected = "".join(data)
    res = obj.reader(mode=r_mode).read()

    if 'b' in r_mode:
        assert res.decode('utf-8') == expected
    else:
        assert res == expected



"""
LenientNamedTuple Module
"""


class LenientNamedTuple:
    """
    Class which provides the NamedTuple functionality but allows initializing with unknown fields.
    The unknown fields will be ignored and mandatory fields are enforced
    """

    __initialized: bool = False
    unknown: dict = None

    def __init__(self, **kwargs):
        fields = list(self.__class__.__dict__["__annotations__"].keys())
        unknown = {}
        for k, v in kwargs.items():
            if k in fields:
                setattr(self, k, v)
                fields.remove(k)
            else:
                unknown[k] = v
        if len(fields) > 0:
            raise TypeError(f"missing {len(fields)} required arguments: {fields}")
        setattr(self, "unknown", unknown)
        self.__initialized = True
        super().__init__()

    def __setattr__(self, name, value):
        if self.__initialized:
            raise AttributeError("can't set attribute")
        super().__setattr__(name, value)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for k, v in self.__dict__.items():
            if k == "unknown":
                continue
            if other.__dict__[k] != v:
                return False

        return True



"""
Module implementing reference's object management logic
"""


class ObjectManager:
    """
    Manage objects listing for a given repository
    """
    # TODO: Implement


class WriteableObjectManager:
    """
    Manage writeable objects for a given repository
    """
    # TODO: Implement



"""
lakeFS Client module


Handles authentication against the lakeFS server and wraps the underlying lakefs_sdk client.
The client module holds a DefaultClient which will attempt to initialize on module loading using
environment credentials.
In case no credentials exist, a call to init() will be required or a Client object must be created explicitly

"""

from __future__ import annotations

from typing import Optional

import lakefs_sdk
from lakefs_sdk.client import LakeFSClient

from lakefs.config import ClientConfig
from lakefs.exceptions import NoAuthenticationFound, NotAuthorizedException, ServerException
from lakefs.namedtuple import LenientNamedTuple

# global default client
DEFAULT_CLIENT: Optional[Client] = None


class ServerStorageConfiguration(LenientNamedTuple):
    """
    Represent a lakeFS server's storage configuration
    """
    blockstore_type: str
    pre_sign_support: bool
    import_support: bool
    blockstore_namespace_example: str
    blockstore_namespace_validity_regex: str
    pre_sign_support_ui: bool
    import_validity_regex: str
    default_namespace_prefix: Optional[str] = None


class ServerConfiguration:
    """
    Represent a lakeFS server's configuration
    """
    _conf: lakefs_sdk.Config
    _storage_conf: ServerStorageConfiguration

    def __init__(self, client: Optional[Client] = DEFAULT_CLIENT):
        try:
            self._conf = client.sdk_client.config_api.get_config()
            self._storage_conf = ServerStorageConfiguration(**self._conf.storage_config.dict())
        except lakefs_sdk.exceptions.ApiException as e:
            if isinstance(e, lakefs_sdk.exceptions.ApiException):
                raise NotAuthorizedException(e.status, e.reason) from e
            raise ServerException(e.status, e.reason) from e

    @property
    def version(self) -> str:
        """
        Return the lakeFS server version
        """
        return self._conf.version_config.version

    @property
    def storage_config(self) -> ServerStorageConfiguration:
        """
        Returns the lakeFS server storage configuration
        """
        return self._storage_conf


class Client:
    """
    Wrapper around lakefs_sdk's client object
    Takes care of instantiating it from the environment
    """

    _client: Optional[LakeFSClient] = None
    _conf: Optional[ClientConfig] = None
    _server_conf: Optional[ServerConfiguration] = None

    def __init__(self, **kwargs):
        self._conf = ClientConfig(**kwargs)
        self._client = LakeFSClient(self._conf.configuration, header_name='X-Lakefs-Client',
                                    header_value='python-lakefs')

    @property
    def config(self):
        """
        Return the underlying lakefs_sdk configuration
        """
        return self._conf.configuration

    @property
    def sdk_client(self):
        """
        Return the underlying lakefs_sdk client
        """
        return self._client

    @property
    def storage_config(self):
        """
        lakeFS SDK storage config object, lazy evaluated.
        """
        if self._server_conf is None:
            self._server_conf = ServerConfiguration(self)
        return self._server_conf.storage_config

    @property
    def version(self) -> str:
        """
        lakeFS Server version, lazy evaluated.
        """
        if self._server_conf is None:
            self._server_conf = ServerConfiguration(self)
        return self._server_conf.version


try:
    DEFAULT_CLIENT = Client()
except NoAuthenticationFound:
    # must call init() explicitly
    DEFAULT_CLIENT = None  # pylint: disable=C0103


def init(**kwargs) -> None:
    """
    Initialize DefaultClient using the provided parameters
    """
    global DEFAULT_CLIENT  # pylint: disable=W0603
    DEFAULT_CLIENT = Client(**kwargs)



"""
Module containing lakeFS branch implementation
"""
from __future__ import annotations

from typing import Optional

import lakefs_sdk

from lakefs.object import WriteableObject
from lakefs.object_manager import WriteableObjectManager
from lakefs.reference import Reference
from lakefs.exceptions import api_exception_handler, ConflictException, LakeFSException


class Branch(Reference):
    """
    Class representing a branch in lakeFS.
    """

    def _get_commit(self):
        """
        For branches override the default _get_commit method to ensure we always fetch the latest head
        """
        self._commit = None
        return super()._get_commit()

    def create(self, source_reference_id: str | Reference, exist_ok: bool = False) -> Branch:
        """
        Create a new branch in lakeFS from this object

        :param source_reference_id: The reference to create the branch from
        :param exist_ok: If False will throw an exception if a branch by this name already exists. Otherwise,
            return the existing branch without creating a new one
        :return: The lakeFS SDK object representing the branch
        :raises:
            NotFoundException if repo, branch or source reference id does not exist
            ConflictException if branch already exists and exist_ok is False
            NotAuthorizedException if user is not authorized to perform this operation
            ServerException for any other errors
        """

        def handle_conflict(e: LakeFSException):
            if isinstance(e, ConflictException) and exist_ok:
                return None
            return e

        branch_creation = lakefs_sdk.BranchCreation(name=self._id, source=str(source_reference_id))
        with api_exception_handler(handle_conflict):
            self._client.sdk_client.branches_api.create_branch(self._repo_id, branch_creation)
        return self

    def head(self) -> Reference:
        """
        Get the commit reference this branch is pointing to

        :return: The commit reference this branch is pointing to
        :raises:
            NotFoundException if branch by this id does not exist
            NotAuthorizedException if user is not authorized to perform this operation
            ServerException for any other errors
        """
        with api_exception_handler():
            branch = self._client.sdk_client.branches_api.get_branch(self._repo_id, self._id)
        return Reference(self._repo_id, branch.commit_id, self._client)

    def commit(self, message: str, metadata: dict = None) -> Reference:
        """
        Commit changes on the current branch

        :param message: Commit message
        :param metadata: Metadata to attach to the commit
        :return: The new reference after the commit
        :raises:
            NotFoundException if branch by this id does not exist
            ForbiddenException if commit is not allowed on this branch
            NotAuthorizedException if user is not authorized to perform this operation
            ServerException for any other errors
        """

        commits_creation = lakefs_sdk.CommitCreation(message=message, metadata=metadata)
        with api_exception_handler():
            c = self._client.sdk_client.commits_api.commit(self._repo_id, self._id, commits_creation)
        return Reference(self._repo_id, c.id, self._client)

    def delete(self) -> None:
        """
        Delete branch from lakeFS server

        :raises:
            NotFoundException if branch or repository do not exist
            NotAuthorizedException if user is not authorized to perform this operation
            ForbiddenException for branches that are protected
            ServerException for any other errors
        """
        with api_exception_handler():
            return self._client.sdk_client.branches_api.delete_branch(self._repo_id, self._id)

    def revert(self, reference_id: str, parent_number: Optional[int] = None) -> None:
        """
        revert the changes done by the provided reference on the current branch

        :param parent_number: when reverting a merge commit, the parent number (starting from 1) relative to which to
            perform the revert.
        :param reference_id: the reference to revert
        :return: The new reference after the revert
        :raises:
            NotFoundException if branch by this id does not exist
            NotAuthorizedException if user is not authorized to perform this operation
            ServerException for any other errors
        """
        if parent_number is None:
            parent_number = 0
        elif parent_number <= 0:
            raise ValueError("parent_number must be a positive integer")

        with api_exception_handler():
            return self._client.sdk_client.branches_api.revert_branch(
                self._repo_id,
                self._id,
                lakefs_sdk.RevertCreation(ref=reference_id, parent_number=parent_number)
            )

    def object(self, path: str) -> WriteableObject:
        """
        Returns a writable object using the current repo id, reference and path

        :param path: The object's path
        """

        return WriteableObject(self.repo_id, self._id, path, client=self._client)

    @property
    def objects(self) -> WriteableObjectManager:
        """
        Return an Objects object using the current repository id and client
        """
        return WriteableObjectManager()



"""
Allow importing of models from package root
"""

from lakefs.repository import Repository, RepositoryProperties
from lakefs.reference import Reference, Commit, Change
from lakefs.tag import Tag
from lakefs.branch import Branch
from lakefs.object import StoredObject, WriteableObject, ObjectReader
from lakefs.branch_manager import BranchManager
from lakefs.object_manager import ObjectManager
from lakefs.tag_manager import TagManager
